#!/usr/bin/env python

# Reference: https://wiki.debian.org/LSBInitScripts

### BEGIN INIT INFO
# Provides:          tinker-access-client
# Required-Start:    $all
# Required-Stop:     $remote_fs $syslog
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: tinker-access-client
# Description:       The tinker-access-client service is responsible for coordinating communication
# between RPi's modules (i.e. RFID reader, LCD, power relays etc..) and the remote tinker_access_server.
### END INIT INFO

import os
import sys
import logging
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientDaemon import ClientDaemon
from CommandHandler import CommandHandler
from ClientOptionParser import ClientOptionParser, Command


def run():
    if os.geteuid() != 0:
        sys.stdout.write(
            'You need to have root privileges to run {0} commands.\n'
            'Please try again, this time using \'sudo\'.\n'.format(PackageInfo.pip_package_name))
        sys.stdout.flush()
        sys.exit(1)

    logger = ClientLogger.setup()
    (opts, args) = ClientOptionParser().parse_args()

    try:
        with CommandHandler(opts, args) as handler:
            handler.on(Command.STOP, ClientDaemon.stop)
            handler.on(Command.START, ClientDaemon.start)
            handler.on(Command.STATUS, ClientDaemon.status)
            handler.on(Command.UPDATE, ClientDaemon.update)
            handler.on(Command.RESTART, ClientDaemon.restart)
            return handler.handle_command()

    except Exception as e:
        logger.exception(e)
        sys.stdout.write(str(e))
        sys.stdout.flush()
        sys.exit(1)

    finally:
        logging.shutdown()

if __name__ == '__main__':
    run()



import os
import cgi
import json
from BaseHTTPServer import BaseHTTPRequestHandler

pins = [
    [
        {'pin': 2, 'desc': '5v Power', 'status': False},
        {'pin': 4, 'desc': '5v Power', 'status': False},
        {'pin': 6, 'desc': 'Ground', 'status': False},
        {'pin': 8, 'desc': 'BCM 14', 'status': True},
        {'pin': 10, 'desc': 'BCM 15', 'status': True},
        {'pin': 12, 'desc': 'BCM 18', 'status': True},
        {'pin': 14, 'desc': 'Ground', 'status': False},
        {'pin': 16, 'desc': 'BCM 23', 'status': True},
        {'pin': 18, 'desc': 'BCM 24', 'status': True},
        {'pin': 20, 'desc': 'Ground', 'status': False},
        {'pin': 22, 'desc': 'BCM 25', 'status': True},
        {'pin': 24, 'desc': 'BCM 8', 'status': True},
        {'pin': 26, 'desc': 'BCM 7', 'status': True},
        {'pin': 28, 'desc': 'BCM 1', 'status': True},
        {'pin': 30, 'desc': 'Ground', 'status': False},
        {'pin': 32, 'desc': 'BCM 12', 'status': True},
        {'pin': 34, 'desc': 'Ground', 'status': False},
        {'pin': 36, 'desc': 'BCM 16', 'status': True},
        {'pin': 38, 'desc': 'BCM 20', 'status': True},
        {'pin': 40, 'desc': 'BCM 21', 'status': True}],
    [
        {'pin': 1, 'desc': '3v3 Power', 'status': False},
        {'pin': 3, 'desc': 'BCM 2', 'status': False},
        {'pin': 5, 'desc': 'BCM 3', 'status': False},
        {'pin': 7, 'desc': 'BCM 4', 'status': False},
        {'pin': 9, 'desc': 'Ground', 'status': False},
        {'pin': 11, 'desc': 'BCM 17', 'status': True},
        {'pin': 13, 'desc': 'BCM 27', 'status': True},
        {'pin': 15, 'desc': 'BCM 22', 'status': True},
        {'pin': 17, 'desc': '3v3 Power', 'status': False},
        {'pin': 19, 'desc': 'BCM 10', 'status': True},
        {'pin': 21, 'desc': 'BCM 9', 'status': True},
        {'pin': 23, 'desc': 'BCM 11', 'status': True},
        {'pin': 25, 'desc': 'Ground', 'status': False},
        {'pin': 27, 'desc': 'BCM 0', 'status': True},
        {'pin': 29, 'desc': 'BCM 5', 'status': True},
        {'pin': 31, 'desc': 'BCM 6', 'status': True},
        {'pin': 33, 'desc': 'BCM 13', 'status': True},
        {'pin': 35, 'desc': 'BCM 19', 'status': True},
        {'pin': 37, 'desc': 'BCM 26', 'status': True},
        {'pin': 39, 'desc': 'Ground', 'status': False}
    ]
]


class VirtualDeviceHandler(BaseHTTPRequestHandler):
    # noinspection PyPep8Naming
    def do_GET(self):
        path = '/VirtualDevice.html' if self.path == '/' else self.path

        try:
            sendReply = False
            if path.endswith(".html"):
                mimetype = 'text/html'
                sendReply = True
            if path.endswith(".jpg"):
                mimetype = 'image/jpg'
                sendReply = True
            if path.endswith(".gif"):
                mimetype = 'image/gif'
                sendReply = True
            if path.endswith(".js"):
                mimetype = 'application/javascript'
                sendReply = True
            if path.endswith(".css"):
                mimetype = 'text/css'
                sendReply = True
            if path.endswith(".json"):
                mimetype = 'application/javascript'
                self.send_response(200)
                self.send_header('Content-type', mimetype)
                self.end_headers()
                self.wfile.write(json.dumps(pins))
                return

            if sendReply:
                f = open('{0}{1}'.format(os.path.dirname(os.path.realpath(__file__)), path))
                self.send_response(200)
                self.send_header('Content-type', mimetype)
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
            return

        except IOError:
            self.send_error(404, 'File Not Found: %s' % self.path)

    # noinspection PyPep8Naming
    def do_POST(self):
        if self.path == "/send":
            form = cgi.FieldStorage(fp=self.rfile, headers=self.headers, environ={
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': self.headers['Content-Type'],
            })

            pin_number = form['pin'].value
            pin_status = form['status'].value == 'false'
            self.update_status(pin_number, pin_status)
            self.send_response(200)
            self.end_headers()
            return

    def update_status(self, pin_number, pin_status):
        for rows in pins:
            for pin in rows:
                if pin['pin'] == int(pin_number):
                    pin['status'] = pin_status
                    break



import os
import sys
import ConfigParser
from optparse import OptionParser, OptionGroup

from Command import Command
from PackageInfo import PackageInfo
from ClientOption import ClientOption

ClientOptionDefaults = {
    ClientOption.DEBUG: False,
    ClientOption.LOG_LEVEL: 40,
    ClientOption.PIN_LOGOUT: 16,
    ClientOption.PIN_LED_RED: 21,
    ClientOption.DEVICE_ID: None,
    ClientOption.PIN_LED_BLUE: 20,
    ClientOption.RESTART_DELAY: 5,
    ClientOption.PIN_LED_GREEN: 19,
    ClientOption.PIN_POWER_RELAY: 17,
    ClientOption.LOGOUT_COAST_TIME: 0,
    ClientOption.PIN_CURRENT_SENSE: 12,
    ClientOption.SERIAL_PORT_SPEED: 9600,
    ClientOption.MAX_POWER_DOWN_TIMEOUT: 5,
    ClientOption.SERIAL_PORT_NAME: '/dev/ttyUSB0',
    ClientOption.SERVER_ADDRESS: 'http://10.2.1.2:5000',
    ClientOption.CONFIG_FILE: '/etc/{0}.conf'.format(PackageInfo.pip_package_name),
    ClientOption.PID_FILE: '/var/run/{0}.pid'.format(PackageInfo.pip_package_name),
    ClientOption.LOG_FILE: '/var/log/{0}.log'.format(PackageInfo.pip_package_name),
    ClientOption.STATUS_FILE: '/var/log/{0}.status'.format(PackageInfo.pip_package_name),
    ClientOption.LOGGING_CONFIG_FILE: '/etc/{0}.logging.conf'.format(PackageInfo.pip_package_name)
}


class ClientOptionParser(object):
    def __init__(self, phase=None):
        self.__phase = phase
        self.__parser = OptionParser(
            version='%prog v{0}'.format(PackageInfo.version) if PackageInfo.version is not None else None)

        if phase == 'install':
            for arg in sys.argv:
                if str(arg).startswith('-'):
                    self.__parser.add_option(str(arg).split('=', 1)[0])

        usage = "\n%prog command [options]"
        commands = ['\n\ncommand:\n']
        for key, value in vars(Command).items():
            if not key.startswith('__'):
                desc = value['description']
                cmd = value['command']
                commands.append('\t%s : %s' % (cmd, desc))
        commands = '\n'.join(commands)
        usage += commands
        usage += '\n\nTinkerMill Raspberry Pi access control system.' \
                 '\n\nExamples:\n\n' \
                 '  Start the client configured to use a different tinker-access-server ' \
                 '(i.e. a development server) and an alternative serial port' \
                 '\n\n  \'sudo {0} --server-address=http://<server-address> ' \
                 '--serial-port-name=/dev/ttyUSB1\' '.format(PackageInfo.python_package_name)

        self.__parser.set_usage(usage)

        self.__parser.add_option(
            '--config-file',
            help='the location of the config file to use [default:\'%default\'] '
                 'a non-default command-line option value will have precedence '
                 'over a config-file option value',
            default=ClientOptionDefaults[ClientOption.CONFIG_FILE],
            dest=ClientOption.CONFIG_FILE,
            action='store')

        self.__parser.add_option(
            '--logging-config-file',
            help='the location of a logging config file to use [default:\'%default\'] '
                 'If this file is present, it will override the default logging configuration '
                 'including the --log-level and --log-file options',
            default=ClientOptionDefaults[ClientOption.LOGGING_CONFIG_FILE],
            dest=ClientOption.LOGGING_CONFIG_FILE,
            action='store')

        self.__parser.add_option(
            '--debug',
            help='run in the foreground (a.k.a debug mode) [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.DEBUG],
            dest=ClientOption.DEBUG,
            action='store_true')

        self.__parser.add_option(
            '--log-file',
            help='the path and name of the log file [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.LOG_FILE],
            dest=ClientOption.LOG_FILE,
            action='store')

        self.__parser.add_option(
            '--status-file',
            help='the path and name of the status file, the contents of this file will always '
                 'reflect the current state of the client. (i.e. initialized, idle, in_use, in_training, terminated) '
                 'A missing file indicates the client is not running [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.STATUS_FILE],
            dest=ClientOption.STATUS_FILE,
            action='store')

        self.__parser.add_option(
            '--pid-file',
            help='the path & name of the pid file [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.PID_FILE],
            dest=ClientOption.PID_FILE,
            action='store')

        self.__parser.add_option(
            '--log-level',
            help='the log level to use [default:%default]',
            default=ClientOptionDefaults[ClientOption.LOG_LEVEL],
            dest=ClientOption.LOG_LEVEL,
            type='int',
            action='store')

        self.__parser.add_option(
            '--server-address',
            help='the api\'s server address [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.SERVER_ADDRESS],
            dest=ClientOption.SERVER_ADDRESS,
            action='store')

        self.__parser.add_option(
            '--device-id',
            help='A unique identity for this client [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.DEVICE_ID],
            dest=ClientOption.DEVICE_ID,
            action='store')

        self.__parser.add_option(
            '--logout-coast-time',
            help='a fixed number of seconds to wait for the physical machine '
                 'to stop after power has been disabled. '
                 '(i.e. a blade to stop spinning etc...) '
                 '[default:%default]',
            default=ClientOptionDefaults[ClientOption.LOGOUT_COAST_TIME],
            dest=ClientOption.LOGOUT_COAST_TIME,
            type='int',
            action='store')

        self.__parser.add_option(
            '--max-power-down-timeout',
            help='the maximum number of seconds to wait for the current sense pin to go low '
                 '[default:%default]',
            default=ClientOptionDefaults[ClientOption.MAX_POWER_DOWN_TIMEOUT],
            dest=ClientOption.MAX_POWER_DOWN_TIMEOUT,
            type='int',
            action='store')

        self.__parser.add_option(
            '--restart-delay',
            help='seconds to wait before attempting to re-start after a failure [default:%default]',
            default=ClientOptionDefaults[ClientOption.RESTART_DELAY],
            dest=ClientOption.RESTART_DELAY,
            type='int',
            action='store')

        gpio_group = OptionGroup(self.__parser, 'RPi GPIO')

        gpio_group.add_option(
            '--pin-logout',
            help='the logout pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_LOGOUT],
            dest=ClientOption.PIN_LOGOUT,
            type='int',
            action='store')

        gpio_group.add_option(
            '--pin-power-relay',
            help='the power relay pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_POWER_RELAY],
            dest=ClientOption.PIN_POWER_RELAY,
            type='int',
            action='store')

        gpio_group.add_option(
            '--pin-led-red',
            help='the red led pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_LED_RED],
            dest=ClientOption.PIN_LED_RED,
            type='int',
            action='store')

        gpio_group.add_option(
            '--pin-led-green',
            help='the green led pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_LED_GREEN],
            dest=ClientOption.PIN_LED_GREEN,
            type='int',
            action='store')

        gpio_group.add_option(
            '--pin-led-blue',
            help='the blue led pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_LED_BLUE],
            dest=ClientOption.PIN_LED_BLUE,
            type='int',
            action='store')

        gpio_group.add_option(
            '--pin-current-sense',
            help='the current sense pin [default:%default]',
            default=ClientOptionDefaults[ClientOption.PIN_CURRENT_SENSE],
            dest=ClientOption.PIN_CURRENT_SENSE,
            type='int',
            action='store')

        self.__parser.add_option_group(gpio_group)

        serial_group = OptionGroup(self.__parser, 'SERIAL')

        serial_group.add_option(
            '--serial-port-name',
            help='the serial port name to use [default:\'%default\']',
            default=ClientOptionDefaults[ClientOption.SERIAL_PORT_NAME],
            dest=ClientOption.SERIAL_PORT_NAME,
            action='store')

        serial_group.add_option(
            '--serial-port-speed',
            help='the serial port speed to use [default:%default]',
            default=ClientOptionDefaults[ClientOption.SERIAL_PORT_SPEED],
            dest=ClientOption.SERIAL_PORT_SPEED,
            type='int',
            action='store')

        self.__parser.add_option_group(serial_group)

    def parse_args(self, args=None, values=None):
        (opts, args) = self.__parser.parse_args(args=args, values=values)
        items = vars(opts)

        options = self.__parser.option_list[:]
        for group in self.__parser.option_groups:
            options = options + group.option_list[:]

        if os.path.isfile(items.get(ClientOption.CONFIG_FILE)):
            config_file_parser = ConfigParser.RawConfigParser()
            config_file_parser.read(items.get(ClientOption.CONFIG_FILE))
            if config_file_parser.has_section('config'):
                for item in config_file_parser.items('config'):
                    option = next((i for i in options if i.dest == item[0]), None)
                    if option:
                        key = item[0]
                        value = item[1]
                        if option.type == 'int':
                            value = int(value)

                        if option.type == 'float':
                            value = float(value)

                        # prevents non-default command-line options from being replaced by config-file options
                        if items.get(key) == option.default != value:
                            items[key] = value

        # TODO: make args[0] required?, check the value and raise error parser.error()
        return items, args



import logging
import webbrowser
from threading import Thread

from tinker_access_client.ClientLogger import ClientLogger
from virtual_device.VirtualDeviceServer import VirtualDeviceServer


class VirtualGPIO(object):
    # https://sourceforge.net/p/raspberry-gpio-python/code/ci/default/tree/source/c_gpio.h

    # define MODE_UNKNOWN -1
    # define BOARD        10
    # define BCM          11
    # define SERIAL       40
    # define SPI          41
    # define I2C          42
    # define PWM          43

    # define SETUP_OK           0
    # define SETUP_DEVMEM_FAIL  1
    # define SETUP_MALLOC_FAIL  2
    # define SETUP_MMAP_FAIL    3
    # define SETUP_CPUINFO_FAIL 4
    # define SETUP_NOT_RPI_FAIL 5

    # define INPUT  1 // is really 0 for control register!
    # define OUTPUT 0 // is really 1 for control register!
    # define ALT0   4

    # define PUD_OFF  0
    HIGH = 1
    LOW = 0
    BCM = 11
    IN = 0
    OUT = 1
    PUD_DOWN = 1
    PUD_UP = 2

    # TODO: confirm these values
    RISING = 1
    FALLING = 2
    BOTH = 3

    def __init__(self):
        self.__logger = logging.getLogger(__name__)
        self.__virtual_device_thread = Thread(target=VirtualDeviceServer.start)
        self.__virtual_device_thread.daemon = True
        self.__virtual_device_thread.start()

    def setmode(self, *args):
        pass

    def cleanup(self, *args):
        pass

    # noinspection PyPep8Naming
    def setwarnings(self, *args):
        pass

    def setup(self, *args):
        pass

    def add_event_detect(self, *args):
        pass

    def add_event_callback(self, *args):
        pass


# # noinspection PyClassHasNoInit
# class VirtualRPi:
#     GPIO = _VirtualGPIO()
#
# GPIO = _VirtualGPIO()



import time
import logging
import threading
from transitions import Machine

from State import State
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from DeviceApi import DeviceApi, Channel
from TinkerAccessServerApi import TinkerAccessServerApi
from ClientOptionParser import ClientOptionParser, ClientOption
from UserRegistrationException import UserRegistrationException
from UnauthorizedAccessException import UnauthorizedAccessException

maximum_lcd_characters = 16
training_mode_delay_seconds = 2
logout_timer_interval_seconds = 1


class Trigger(object):
    IDLE = 'idle'
    LOGIN = 'login'
    LOGOUT = 'logout'
    TERMINATE = 'terminate'


# noinspection PyUnusedLocal
class Client(Machine):
    def __init__(self, device=None):
        self.__device = device
        self.__user_info = None
        self.__logout_timer = None
        self.__logger = ClientLogger.setup()
        self.__opts = ClientOptionParser().parse_args()[0]
        self.__tinkerAccessServerApi = TinkerAccessServerApi(self.__opts)

        states = []
        for key, value in vars(State).items():
            if not key.startswith('__'):
                states.append(value)

        transitions = [
            {
                'source': [State.INITIALIZED],
                'trigger': Trigger.IDLE,
                'dest': State.IDLE
            },

            {
                'source': [State.IDLE],
                'trigger': Trigger.LOGIN,
                'dest': State.IN_USE,
                'conditions': ['is_authorized']
            },

            {
                'source': [State.IN_USE],
                'trigger': Trigger.LOGIN,
                'dest': State.IN_USE,
                'conditions': ['should_extend_current_session']
            },

            {
                'source': [State.IN_USE, State.IN_TRAINING],
                'trigger': Trigger.LOGOUT,
                'dest': State.IDLE
            },

            {
                'source': [State.IDLE],
                'trigger': Trigger.LOGOUT,
                'dest': State.IN_TRAINING,
                'conditions': ['is_waiting_for_training']
            },

            {
                'source': '*',
                'trigger': Trigger.TERMINATE,
                'dest': State.TERMINATED
            }
        ]

        Machine.__init__(self, queued=True, states=states,
                         transitions=transitions, initial=State.INITIALIZED, after_state_change='update_status')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    #
    # idle -- The machine is idle and waiting for a badge to be scanned
    #

    def __ensure_idle(self):
        self.__do_logout()
        self.__disable_power()
        self.__show_blue_led()
        self.__show_scan_badge()

    def __do_login(self, *args, **kwargs):
        badge_code = kwargs.get('badge_code')

        # noinspection PyBroadException
        try:
            self.__show_attempting_login(1)
            self.__update_user_context(
                self.__tinkerAccessServerApi.login(badge_code)
            )
            remaining_seconds = self.__user_info.get('remaining_seconds')
            self.__show_access_granted(1)

        except UnauthorizedAccessException as e:
            self.__handle_unauthorized_access_exception()
            self.__ensure_idle()
        except Exception as e:
            self.__handle_unexpected_exception()
            self.__ensure_idle()
            raise e

        return self.__user_info is not None

    def __show_attempting_login(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ATTEMPTING'.center(maximum_lcd_characters, ' '),
            'LOGIN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __update_user_context(self, user_info):
        self.__user_info = user_info
        for handler in self.__logger.handlers:
            for context_filter in handler.filters:
                update_user_context = getattr(context_filter, "update_user_context", None)
                if callable(update_user_context):
                    context_filter.update_user_context(self.__user_info)

    def __handle_unauthorized_access_exception(self):
        self.__show_red_led()
        self.__show_access_denied(2)

    def __show_access_denied(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ACCESS DENIED'.center(maximum_lcd_characters, ' '),
            'TAKE THE CLASS'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_red_led(self):
        self.__device.write(Channel.LED, True, False, False)

    def __do_logout(self):
        self.__cancel_logout_timer()

        if self.__user_info:
            badge_code = self.__user_info.get('badge_code')
            threading.Timer(0, self.__tinkerAccessServerApi.logout, (badge_code, )).start()

        self.__update_user_context(None)

    def __show_blue_led(self):
        self.__device.write(Channel.LED, False, False, True)

    def __show_scan_badge(self):
        self.__device.write(
            Channel.LCD,
            'SCAN BADGE'.center(maximum_lcd_characters, ' '),
            'TO LOGIN'.center(maximum_lcd_characters, ' ')
        )

    #
    # in_use -- The machine is currently in use and the logout timer is ticking...
    #

    def __ensure_in_use(self):
        self.__enable_power()
        self.__show_green_led()

    def __enable_power(self):
        self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_POWER_RELAY), True)

    def __show_access_granted(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ACCESS GRANTED'.center(maximum_lcd_characters, ' '),
        )
        time.sleep(delay)

    def __show_green_led(self):
        self.__device.write(Channel.LED, False, True, False)

    def __start_logout_timer(self):
        self.__cancel_logout_timer()
        self.__logout_timer = threading.Timer(
            logout_timer_interval_seconds,
            self.__logout_timer_tick
        )
        self.__logout_timer.start()

    def __logout_timer_tick(self):
        if not self.is_terminated() and self.__user_info:
            remaining_seconds = self.__user_info.get('remaining_seconds')

            if remaining_seconds <= 0:
                self.logout()
                return

            self.__user_info['remaining_seconds'] = (remaining_seconds - logout_timer_interval_seconds)
            self.__show_remaining_time()
            self.__start_logout_timer()

    def __show_remaining_time(self):
        remaining_seconds = self.__user_info.get('remaining_seconds')
        if remaining_seconds < 300:
            self.__toggle_red_led()

        m, s = divmod(int(remaining_seconds), 60)
        h, m = divmod(m, 60)
        user_name = self.__user_info.get('user_name')
        self.__device.write(
            Channel.LCD,
            user_name.center(maximum_lcd_characters, ' '),
            '{0:02d}:{1:02d}:{2:02d}'.format(h, m, s).center(maximum_lcd_characters, ' ')
        )

    def __toggle_red_led(self):
        red_led_status = self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LED_RED))
        self.__device.write(Channel.LED, not red_led_status, False, False)

    def __cancel_logout_timer(self):
        if self.__logout_timer:
            self.__logout_timer.cancel()

        self.__logout_timer = None

    def __extend_session(self):
        self.__cancel_logout_timer()

        # TODO: add api call to let server know that time has been extended...

        session_seconds = self.__user_info.get('session_seconds')
        remaining_seconds = self.__user_info.get('remaining_seconds')
        remaining_extensions = self.__user_info.get('remaining_extensions')

        if remaining_extensions:
            if remaining_extensions != float('inf'):
                self.__user_info['remaining_extensions'] = remaining_extensions - 1

            remaining_seconds = remaining_seconds + session_seconds
            self.__user_info['remaining_seconds'] = remaining_seconds
            self.__logger.info('Session extended %s remaining_seconds', remaining_seconds)

            self.__show_session_extended(1)
        else:
            self.__show_no_extensions_remaining(2)

        self.__start_logout_timer()

    def __show_session_extended(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'SESSION EXTENDED'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        self.__show_remaining_time()

    def __show_no_extensions_remaining(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'NO EXTENSIONS'.center(maximum_lcd_characters, ' '),
            'REMAINING...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        self.__show_remaining_time()

    #
    # training - the client has entered training mode
    #

    def __show_training_mode_activated(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'TRAINING MODE'.center(maximum_lcd_characters, ' '),
            'ACTIVATED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __activate_trainer(self, badge_code):
        # Note: currently we call the normal login method on the tinkerAccessServer
        # the backend ensures the user is a trainer on the 'registerUser' call.
        # I expect this might change in the future once the server code gets refactored
        try:
            self.__show_attempting_login(1)
            self.__update_user_context(
                self.__tinkerAccessServerApi.login(badge_code)
            )
            self.__show_trainer_accepted(1)
            self.__prompt_for_student_badge()

        except UnauthorizedAccessException:
            self.__handle_unauthorized_access_exception()
            self.__prompt_for_trainer_badge()
        except Exception as e:
            self.__handle_unexpected_exception()
            self.__prompt_for_trainer_badge()
            raise e

        return self.__user_info is not None

    def __prompt_for_trainer_badge(self):
        self.__show_blue_led()
        self.__show_scan_trainer_badge()

    def __show_scan_trainer_badge(self):
        self.__device.write(
            Channel.LCD,
            'SCAN'.center(maximum_lcd_characters, ' '),
            'TRAINER BADGE...'.center(maximum_lcd_characters, ' ')
        )

    def __show_trainer_accepted(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'TRAINER'.center(maximum_lcd_characters, ' '),
            'ACCEPTED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __prompt_for_student_badge(self):
        self.__show_blue_led()
        self.__show_scan_student_badge()

    def __show_scan_student_badge(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'SCAN'.center(maximum_lcd_characters, ' '),
            'STUDENT BADGE...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __register_student(self, badge_code):
        try:
            self.__show_attempting_registration(1)
            trainer_id = self.__user_info.get('user_id')
            trainer_badge_code = self.__user_info.get('badge_code')
            self.__tinkerAccessServerApi.register_user(trainer_id, trainer_badge_code, badge_code)
            self.__show_student_registered(1)
        except UserRegistrationException:
            self.__handle_user_registration_exception()
            self.__show_unknown_user(2)
        except Exception as e:
            self.__handle_user_registration_exception()
            self.__handle_unexpected_exception()
            raise e
        finally:
            self.__prompt_for_student_badge()

    def __show_attempting_registration(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ATTEMPTING'.center(maximum_lcd_characters, ' '),
            'REGISTRATION...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_student_registered(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'STUDENT'.center(maximum_lcd_characters, ' '),
            'REGISTERED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __handle_user_registration_exception(self):
        self.__show_red_led()
        self.__show_registration_failed(2)

    def __show_registration_failed(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'REGISTRATION'.center(maximum_lcd_characters, ' '),
            'FAILED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_unknown_user(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'UNKNOWN'.center(maximum_lcd_characters, ' '),
            'USER...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
    #
    # logout/terminated - the user has logged out, or the client is shutting down
    #

    def __disable_power(self):
        self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_POWER_RELAY), False)
        self.__wait_for_power_down()
        self.__wait_for_logout_coast_time()

    def __wait_for_power_down(self):
        current_sense_pin = self.__opts.get(ClientOption.PIN_CURRENT_SENSE)
        max_power_down_timeout = self.__opts.get(ClientOption.MAX_POWER_DOWN_TIMEOUT)

        current = time.time()
        while time.time() - current < max_power_down_timeout and self.__device.read(Channel.PIN, current_sense_pin):
            self.__show_disabling_power()
            time.sleep(0.1)

    def __wait_for_logout_coast_time(self):
        logout_coast_time = self.__opts.get(ClientOption.LOGOUT_COAST_TIME)
        if logout_coast_time:
            self.__show_coasting_down()
            time.sleep(logout_coast_time)

    def __show_disabling_power(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'DISABLING'.center(maximum_lcd_characters, ' '),
            'POWER...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_coasting_down(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'COASTING'.center(maximum_lcd_characters, ' '),
            'DOWN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __handle_unexpected_exception(self):
        self.__show_red_led()
        self.__show_error_occurred(2)
        self.__show_please_try_again(2)

    def __show_error_occurred(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'THERE WAS AN'.center(maximum_lcd_characters, ' '),
            'UNEXPECTED ERROR'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_please_try_again(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'PLEASE'.center(maximum_lcd_characters, ' '),
            'TRY AGAIN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
    #
    # a badge code has been detected on the serial input
    #

    def handle_badge_code(self, *args, **kwargs):
        badge_code = kwargs.get('badge_code')
        if self.state is State.IN_TRAINING:
            if not self.__user_info:
                if self.__activate_trainer(badge_code):
                    self.__show_scan_student_badge()
                else:
                    self.__show_scan_trainer_badge()
            else:
                if not self.__is_current_badge_code(*args, **kwargs):
                    self.__register_student(badge_code)
        else:
            self.login(*args, **kwargs)


    #
    # wait - wait for the next edge detection event from the device.
    #

    def wait(self):
        self.__device.wait()

    #
    # status - The client has received a status command
    #

    def status(self):
        return self.state

    def update_status(self, *args, **kwargs):
        status_file = self.__opts.get(ClientOption.STATUS_FILE)
        with open(status_file, 'w') as f:
            f.write('{0}\n'.format(self.status()))
    #
    # conditions - used to allow/prevent triggers causing a transition if the conditions are not met.
    #

    def is_terminated(self):
        return self.state is State.TERMINATED

    def is_authorized(self, *args, **kwargs):
        return self.__do_login(*args, **kwargs)

    def is_waiting_for_training(self, *args, **kwargs):
        current = time.time()
        while not self.is_terminated() and time.time() - current < training_mode_delay_seconds \
                and self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LOGOUT)):
            time.sleep(0.1)

        return self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LOGOUT))

    def should_extend_current_session(self, *args, **kwargs):
        if self.__is_current_badge_code(*args, **kwargs):
            self.__extend_session()
            return True
        return False

    def __is_current_badge_code(self, *args, **kwargs):
        new_badge_code = kwargs.get('badge_code')
        current_badge_code = self.__user_info.get('badge_code') if self.__user_info else None

        if current_badge_code and current_badge_code == new_badge_code:
            return True

        return False

    def on_enter_idle(self, *args, **kwargs):
        self.__ensure_idle()
        self.__show_scan_badge()

    def on_enter_in_use(self, *args, **kwargs):
        self.__ensure_in_use()
        self.__start_logout_timer()

    def on_enter_in_training(self, *args, **kwargs):
        self.__ensure_idle()
        self.__show_training_mode_activated(1)
        self.__show_scan_trainer_badge()

    def on_enter_terminated(self, *args, **kwargs):
        self.__ensure_idle()

    @staticmethod
    def run(opts, args):
        should_exit = False
        logger = ClientLogger.setup()
        restart_delay = opts.get(ClientOption.RESTART_DELAY)

        while not should_exit:
            try:
                with DeviceApi(opts) as device, Client(device) as client:

                    device.on(
                        Channel.SERIAL,
                        direction=device.GPIO.IN,
                        call_back=client.handle_badge_code
                    )

                    device.on(
                        Channel.PIN,
                        pin=opts.get(ClientOption.PIN_LOGOUT),
                        direction=device.GPIO.RISING,
                        call_back=client.logout
                    )

                    client.idle()
                    while not client.is_terminated():
                        logger.debug('%s is waiting...', PackageInfo.pip_package_name)
                        client.wait()
                    should_exit = True

            except (KeyboardInterrupt, SystemExit):
                should_exit = True

            except Exception as e:
                logger.exception(e)
                logger.debug('Retrying in %s seconds...', restart_delay)
                time.sleep(restart_delay)

        logging.shutdown()



from setuptools import setup, find_packages
from setuptools.command.install import install as _install

from tinker_access_client.ServiceInstaller import ServiceInstaller
from tinker_access_client.PackageInfo import PackageInfo


# noinspection PyClassHasNoInit,PyPep8Naming
class install(_install):
    def run(self):
        _install.run(self)
        msg = '\nInstalling {0} service...'.format(PackageInfo.python_package_name)
        self.execute(ServiceInstaller(self.install_lib).install, [], msg)


config = {
    'name': PackageInfo.pip_package_name,
    'description': PackageInfo.pip_package_name,
    'author': 'Erick McQueen',
    'url': 'http://github.com/tinkerMill/tinkerAccess',
    'download_url': 'https://github.com/tinkerAccess/archive/v{0}.tar.gz'.format(PackageInfo.version),
    'author_email': 'ronn.mcqueen@tinkermill.org',
    'version': PackageInfo.version,
    'zip_safe': False,
    'include_package_data': True,
    'package_data': {
        PackageInfo.python_package_name: [
            'scripts/*'
        ]
    },
    'install_requires': [
        'transitions==0.4.3',
        'daemonize==2.4.7',
        'requests==2.12.4',
        'pyserial==3.2.1',
        'retry==0.9.2'
    ],

    'packages': find_packages(exclude=('tests*',)),
    'test_suite': 'nose.collector',
    'tests_require': [
        'transitions==0.4.3',
        'daemonize==2.4.7',
        'requests==2.12.4',
        'pyserial==3.2.1',
        'mock==2.0.0',
        'nose==1.3.7'
    ],
    'entry_points': {
        'console_scripts': [
            '{0}={1}.Service:run'.format(PackageInfo.pip_package_name, PackageInfo.python_package_name)
        ]
    },
    'cmdclass': {
        'install': install
    }
}

setup(**config)
